// TODO : Autojoin docker swarm
// TODO : Autojoin teleport
// TODO : Install CA certs
// TODO : Install node exporter
// TODO : ZFS?

// DRIVE SELECTION

use crate::config::InstallMode;
use crate::fx::docker::setup_docker;
use crate::fx::firewall::setup_firewall;
use crate::fx::gpu::setup_video_drivers;
use crate::fx::ollama::setup_ollama;
use crate::fx::ssh::setup_ssh;
use crate::fx::virt::{VirtFeature, setup_virtualization, try_setup_vm_guest};
use crate::fx::zram::setup_zram;
use boot::setup_bootloader;
use desktop::setup_desktop;
use drives::{format_drives, mount_drives, setup_fstrim};
use firmware::{setup_fwupd, setup_microcode};
use first_boot::{first_boot_values, genfstab};
use kernel::setup_mkinitcpio;
use navos::setup_navos;
use security::{has_secure_boot, setup_secure_boot, setup_tpm_unlock};
use skel::setup_skel;
use user::setup_users;
use yansi::{Color, Paint};

pub mod boot;
pub mod desktop;
pub mod drives;
pub mod firmware;
pub mod first_boot;
pub mod kernel;
pub mod navos;
pub mod security;
pub mod skel;
pub mod user;

// TODO : error handling
// TODO : Power profile daemon
// TODO : Make zsh default + completions

use crate::{
    config::{InstallConfig, WlanConfig},
    fx::bluetooth::setup_bluetooth,
    linux::run_command,
    pkg::{self, install_pkgs, pacstrap_at, setup_kernel},
    print_status,
};

/// Uncomment the first occurrence of a specified value in a file.
///
/// This function searches for the first line in the specified file that contains
/// the given `value` string. If the line is commented out with a `#` symbol,
/// it removes the `#` and updates the file.
///
/// # Arguments
///
/// * `value` - A string that specifies the value to search for in the file.
/// * `file` - A string specifying the path to the file.
pub fn uncomment_first_value_of(value: &str, file: &str) {
    // read in the file
    let content = std::fs::read_to_string(file).unwrap();
    let mut new = String::new();
    let mut found = false;

    // search for the first instance of `value` in the file
    // uncomment the '#' symbol if there is one
    for line in content.lines() {
        if line.contains(value) && !found {
            new.push_str(&format!("{}\n", line.replace('#', "")));
            found = true;
        } else {
            new.push_str(&format!("{line}\n"));
        }
    }
    // write back
    std::fs::write(file, new).unwrap();
}

/// Uncomment lines that start with a specified tag in a file.
///
/// This function processes all lines in the specified file. For any line that starts
/// with the provided `tag`, it removes the `tag` and updates the file.
///
/// # Arguments
///
/// * `tag` - A string specifying the prefix tag to remove from matching lines.
/// * `file` - A string specifying the path to the file.
pub fn uncomment_tag(tag: &str, file: &str) {
    // read in the file
    let content = std::fs::read_to_string(file).unwrap();
    let mut new = String::new();

    // search for the first instance of `value` in the file
    // uncomment the '#' symbol if there is one
    for line in content.lines() {
        if line.starts_with(tag) {
            new.push_str(&format!("{}\n", line.trim_start_matches(tag)));
        } else {
            new.push_str(&format!("{line}\n"));
        }
    }

    // write back
    std::fs::write(file, new).unwrap();
}

/// Install a config on a new system
pub fn install(mut conf: InstallConfig, bare: bool) {
    if conf.drive.disk.is_some() && (conf.drive.boot.is_some() || conf.drive.root.is_some()) {
        println!("The `drive` option cannot be used with `root` / `boot` options")
    }

    conf.drive.use_entire_disk();

    // Drive Setup
    format_drives(&conf.drive);
    mount_drives(&conf.drive);
    install_mnt(conf, bare);
}

pub fn setup_pacman(dir: &str, conf: &str) {
    print_status("Installing pacman.conf");
    let dir = std::path::Path::new(dir);
    let content = std::fs::read_to_string(&conf).unwrap();
    std::fs::write(dir.join("etc").join("pacman.conf"), content).unwrap();
}

pub fn setup_wlan(wlan: &WlanConfig) {
    print_status("Setting base wlan");
    install_pkgs(&["networkmanager"]);
    let res = run_command(
        &[
            "wpa_passphrase",
            wlan.ssid.as_str(),
            wlan.passphrase.as_str(),
        ],
        None,
        false,
    )
    .0;
    let psk = res
        .lines()
        .find(|x| x.trim().starts_with("psk"))
        .map(|x| x.trim_start_matches("psk=").to_string())
        .unwrap();
    std::fs::write(
        "/mnt/etc/NetworkManager/system-connections/base.nmconnection",
        include_str!("../root/wlan.nmconnection")
            .replace("<PSK>", &psk)
            .replace("<SSID>", &wlan.ssid),
    )
    .unwrap();
}

pub fn install_mnt(conf: InstallConfig, bare: bool) {
    // Base Install
    pacstrap_at("/mnt", &conf.pkg.pkg);
    setup_pacman(
        "/mnt",
        &conf
            .pkg
            .pacman
            .clone()
            .unwrap_or("/etc/pacman.conf".to_string()),
    );
    setup_kernel(conf.pkg.kernel.clone());
    genfstab();

    // Configuration
    first_boot_values(&conf.general);

    // Setup network
    if let Some(wlan) = &conf.wlan {
        setup_wlan(wlan);
    }

    if !matches!(conf.general.mode, crate::config::InstallMode::Base) {
        setup_navos("/mnt");
    }

    setup_skel(&conf.general);
    setup_users(&conf.user.as_ref().unwrap_or(&Vec::new()));
    setup_ssh(&conf.ssh);

    // Presets
    match conf.general.mode {
        crate::config::InstallMode::Base => {}
        crate::config::InstallMode::Desktop => {
            setup_desktop(&conf);
        }
        crate::config::InstallMode::Server => {
            install_pkgs(&pkg::SERVER_PKG);
        }
        crate::config::InstallMode::Kiosk => {
            // TODO
        }
    }

    // Applications

    if conf.pkg.virtualization.unwrap_or_default() {
        setup_virtualization(&VirtFeature {
            gui: matches!(conf.general.mode, InstallMode::Desktop),
            user: conf
                .user
                .as_ref()
                .map(|users| {
                    users
                        .into_iter()
                        .filter_map(|x| {
                            if x.virtualization.unwrap_or_default() {
                                Some(x.name.clone())
                            } else {
                                None
                            }
                        })
                        .collect::<Vec<_>>()
                })
                .unwrap_or_default(),
        });
    }

    if conf.pkg.docker.unwrap_or_default() {
        let user_conf = if let Some(user_conf) = &conf.user {
            user_conf.clone()
        } else {
            Vec::new()
        };

        let docker_users: Vec<_> = user_conf
            .iter()
            .filter_map(|x| {
                if x.docker.unwrap_or_default() {
                    Some(x.name.clone())
                } else {
                    None
                }
            })
            .collect();
        setup_docker(&docker_users);
    }

    if let Some(ai) = conf.ai {
        setup_ollama(&ai);
    }

    // Connectivity
    if conf.general.bluetooth.unwrap_or_default() {
        setup_bluetooth();
    }

    if conf.general.firewall.unwrap_or(true) {
        setup_firewall(conf.ssh.is_some());
    }

    // System
    setup_zram();
    try_setup_vm_guest();
    if let Some(gpu_vendor) = &conf.general.gpu_driver {
        setup_video_drivers(gpu_vendor);
    }
    setup_fwupd();
    setup_microcode();
    setup_fstrim();
    setup_bootloader();
    setup_mkinitcpio(&conf.drive, conf.general.secure_boot.unwrap_or(false));

    if bare && has_secure_boot() {
        if conf.general.secure_boot.unwrap_or(true) {
            setup_secure_boot();
        } else {
            print_status("Skipping Secure Boot");
        }
    }

    if conf.drive.encryption.is_some() {
        setup_tpm_unlock(&conf.drive);
    }

    println!("{}", "System install complete".bold().paint(Color::Green));
}
