// TPM Unlock

use yansi::{Color, Paint};

use crate::{
    config::DriveConfig,
    linux::{arch_chroot, install_file, run_command_noerr, systemd_service_enable},
    pkg::install_pkgs,
};

/// Setup TPM Unlock and a recovery key for the root drive
pub fn setup_tpm_unlock(conf: &DriveConfig) {
    install_pkgs(&["tpm2-tools"]);

    // systemd-cryptenroll --tpm2-device=list

    // Recovery Key
    let recovery_key = arch_chroot(
        &vec![
            "systemd-run",
            "--pipe",
            "--collect",
            "-p",
            format!(
                "SetCredential=cryptenroll.passphrase:{}",
                conf.encryption.as_ref().unwrap()
            )
            .as_str(),
            "systemd-cryptenroll",
            "--recovery-key",
            &conf.root.as_ref().unwrap(),
        ],
        None,
        false,
    )
    .0;

    install_file("/mnt/root/recovery.key", &recovery_key, 0o400);

    install_file(
        "/mnt/etc/systemd/system/tpm-enroll.service",
        &include_str!("../root/tpm-enroll.service")
            .replace("<PASSPHRASE>", conf.encryption.as_ref().unwrap())
            .replace("<ROOT>", &conf.root.as_ref().unwrap()),
        0o644,
    );
    systemd_service_enable("tpm-enroll.service");
}

// SECURE BOOT

pub fn has_secure_boot() -> bool {
    let (stdout, _) = run_command_noerr(&["sbctl", "status"], None, false);
    let binding = stdout.lines().collect::<Vec<&str>>();
    if let Some(status) = binding.get(1) {
        if status.contains("Setup Mode") || status.contains("Enabled") {
            return true;
        }
    }

    false
}

pub fn ensure_secure_boot() {
    if !has_secure_boot() {
        println!(
            "{}",
            "[!] Secure Boot is not in Setup Mode".paint(Color::Red)
        );
        std::process::exit(1);
    }
}

/// Setup Secure Boot on the system
pub fn setup_secure_boot() {
    ensure_secure_boot();

    install_pkgs(&["sbctl", "sbsigntools"]);

    arch_chroot(&["sbctl", "create-keys"], None, false);

    arch_chroot(&["sbctl", "enroll-keys", "--microsoft"], None, false);

    sb_sign("/boot/EFI/Linux/arch-linux.efi");
    sb_sign("/boot/EFI/Linux/arch-linux-fallback.efi");
    sb_sign("/boot/EFI/systemd/systemd-bootx64.efi");
    sb_sign("/boot/EFI/Boot/bootx64.efi");

    arch_chroot(&["sbctl", "verify"], None, false);
}

pub fn sb_sign(file: &str) {
    arch_chroot(&["sbctl", "sign", "-s", file], None, false);
}
